# import asyncio
# import os
#
# from loguru import logger
# from sanic import Sanic, Request, json
#
# from web_foundation import settings
# from web_foundation.environment.events.settings import SettingsChange
# from web_foundation.kernel.messaging.channel import IChannel
# from web_foundation.kernel.messaging.dispatcher import IDispatcher
#
# settings.DEBUG = True
# settings.METRICS_ENABLE = True
# settings.EVENTS_METRIC_ENABLE = True
#
# app = Sanic("name")
# app.shared_ctx.disp = IDispatcher()
#
#
# async def call_event(event):
#     logger.warning(f"new event {os.getpid()} {event}")
#
#
# def listen_disp(disp: IDispatcher):
#     async def listen(disp):
#         await asyncio.gather(*disp.perform())
#
#     try:
#         asyncio.run(listen(disp))
#     except KeyboardInterrupt:
#         print("done")
#
#
# @app.before_server_start
# async def find_channel(s_app: Sanic):
#     worker_num = int(s_app.m.name.split('-')[-2])
#     channel = getattr(s_app.shared_ctx, f"Channel_{worker_num}")
#     s_app.ctx.channel = channel
#     s_app.ctx.channel.add_event_listener(SettingsChange, call_event)
#
#
# @app.after_server_start
# async def listen(s_app: Sanic):
#     await s_app.ctx.channel.listen_consume()
#
#
# @app.main_process_start
# async def start(s_app: Sanic):
#     for i in range(s_app.state.workers):
#         channel = IChannel(f"Channel_{i}")
#         setattr(s_app.shared_ctx, f"Channel_{i}", channel)
#         s_app.shared_ctx.disp.channels.update({f"Channel_{i}": channel})
#
#
# #
# @app.main_process_ready
# async def ready(s_app: Sanic):
#     s_app.manager.manage("IDispatcher", listen_disp, {"disp": s_app.shared_ctx.disp})  # TODO KeyboardInterrupt signal
#
#
# @app.get("/test")
# async def test(r: Request):
#     logger.warning(f"{os.getpid()}")
#     # logger.warning(r.app.shared_ctx)
#     data = SettingsChange("DEBUG", settings.DEBUG)
#     logger.warning(r.app.m.name)
#     # logger.warning(r.app.ctx.channel.produce_pipe.queue.put(data))
#     await r.app.ctx.channel.produce(data)
#     # r.app.shared_ctx.q: Queue
#     # r.app.shared_ctx.q.put(1, block=False)
#     return json({"ok": "ok"})
#
#
# if __name__ == '__main__':
#     app.run("0.0.0.0", fast=True,dev=True)
#
# """
#  ctx:
#     Channel
#     AddonLoader
#     DepContainer:
#         RTManager
#         AddonService
#         MetricService
#         *Services
#         *Resource
# """
from enum import Enum
from pathlib import Path

if __name__ == '__main__':
    print(Path("../applied_files").joinpath(Path("config")).joinpath(Path("")))



from abc import ABCMeta
from typing import Any

from web_foundation.resources.resource import Resource


class FilesResource(Resource, metaclass=ABCMeta):
    root: Any

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    async def open(self, path: Any, *args, **kwargs) -> Any:
        raise NotImplementedError

    async def remove(self, path: Any, recursive: bool = False, *args, **kwargs):
        raise NotImplementedError

    async def rename(self, path: Any, new_name: str, *args, **kwargs):
        raise NotImplementedError

    async def replace(self, path: Any, new_path: str, *args, **kwargs):
        raise NotImplementedError

    async def mkdir(self, path: Any, *args, **kwargs):
        raise NotImplementedError

    async def makedirs(self, path: Any, *args, **kwargs):
        raise NotImplementedError

    async def list(self, path: Any, *args, **kwargs):
        raise NotImplementedError

    async def exists(self, path: Any) -> bool:
        raise NotImplementedError



from __future__ import annotations

from typing import TypeVar

from pydantic import BaseModel

from web_foundation.config import GenericConfig
from web_foundation.extentions.addons.addon_loader import AddonsLoader
from web_foundation.kernel.dependency import Dependency
from web_foundation.kernel.messaging.channel import IChannel
from web_foundation.kernel.service import Service
from web_foundation.kernel.stage import ListenerRegistrator
from web_foundation.resources.files.interface import FilesResource
from web_foundation.resources.resource import Resource
from web_foundation.resources.stores.interface import DataStore
from web_foundation.utils.helpers import in_obj_subclasses, in_obj_classes, in_obj_subcls_of_dependencies


class DependencyContainer:
    app_config: Dependency[GenericConfig] = Dependency(instance_of=BaseModel)
    data_store: Dependency[DataStore] = Dependency(instance_of=DataStore)
    file_repository: Dependency[FilesResource] = Dependency(instance_of=FilesResource)
    addon_loader: Dependency[AddonsLoader] = Dependency(instance_of=AddonsLoader)

    def __init__(self, **kwargs):
        for dep, name in in_obj_classes(self, Dependency):
            dep_init = kwargs.get(name)
            if dep_init:
                dep.from_initialized(dep_init)

    @property
    def resources(self):
        for resource, name in in_obj_subcls_of_dependencies(self, Resource):
            if isinstance(resource, Dependency):
                resource = resource()
            yield resource

    async def before_app_run(self):
        pass

    def set_listeners(self, registrator: ListenerRegistrator):
        for dep, name in in_obj_classes(self, Dependency):
            if inited := dep():
                if hasattr(inited, "set_listeners"):
                    inited.set_listeners(registrator)

    async def init_resources(self, channel: IChannel | None = None):
        for resource in self.resources:
            await resource.init(self.app_config(), channel)

    async def shutdown_resources(self):
        for resource in self.resources:
            await resource.shutdown()

    def services(self):
        return set(in_obj_subclasses(self, Service))

GenericDependencyContainer = TypeVar("GenericDependencyContainer", bound=DependencyContainer)



from pathlib import Path
from typing import TypeVar

from pydantic import BaseSettings, BaseModel


class DbConfig(BaseModel):
    host: str
    port: str
    database: str
    user: str
    password: str
    db_schema: str
    with_migrations: bool
    migrations_path: Path


class SanicConf(BaseModel):
    host: str | None
    port: int | None
    dev: bool = False
    fast: bool = False
    debug: bool = False
    verbosity: int = 0
    workers: int = 1
    access_log: bool | None
    noisy_exceptions: bool | None

    # TODO:MABY
    # auto_reload: bool | None
    # reload_dir: Optional[Union[List[str], str]] = None


class AppConf(BaseModel):
    app_name: str
    database: DbConfig | None
    sanic: SanicConf | None


GenericConfig = TypeVar("GenericConfig", bound=BaseModel)



from abc import ABCMeta


class AbstractModule(metaclass=ABCMeta):
    _available: bool

    def __init__(self, available: bool):
        self._available = available

    async def emit(self):
        print("event_emitted")

    @property
    def available(self):
        return self._available



from typing import TypeVar, Dict

import loguru
from apscheduler.triggers.base import BaseTrigger

from web_foundation.kernel.messaging.channel import IMessage, IChannel
from web_foundation.extentions.schedulle.scheduller import TaskIMessage, BackgroundTask, TaskErrorCallback


class Service:
    channel: IChannel

    async def run_background(self, task: BackgroundTask,
                             *args,
                             trigger: BaseTrigger = None,
                             on_error_callback: TaskErrorCallback = None,
                             add_job_kw: Dict = None, return_event: bool = False, **kwargs,
                             ):
        await self.emmit_event(
            TaskIMessage(task,
                         trigger,
                         on_error_callback=on_error_callback,
                         add_job_kw=add_job_kw,
                         args=args,
                         kwargs=kwargs,
                         return_event=return_event))

    async def wait_for_response(self, msg: IMessage):
        return await self.channel.produce_for_response(msg)

    async def emmit_event(self, event: IMessage):
        await self.channel.produce(event)


GenericService = TypeVar("GenericService", bound=Service)



import asyncio

from web_foundation.extentions.metrics.events import MetricRequestEvent, NewMetricEvent, MetricResponseEvent
from web_foundation.extentions.metrics.exporter import MetricExporter
from web_foundation.extentions.metrics.metric import Metric
from web_foundation.kernel.service import Service


class MetricsService(Service):
    async def collect_metrics(self, exporter: MetricExporter, kind: str = "__system_metric__"):
        metrics_response: MetricResponseEvent = await self.wait_for_response(MetricRequestEvent(exporter, kind))
        return metrics_response.metrics_data

    async def give_metric(self, metric: Metric, kind: str = "__system_metric__"):
        metric.kind = kind
        asyncio.create_task(self.channel.produce(NewMetricEvent(metric)))



from __future__ import annotations

from typing import Generic, TypeVar
from typing import Type

T = TypeVar("T")


class Dependency(Generic[T]):  # type: ignore
    _initialized: T | None = None
    default: T | None
    instance_of: Type[T] | None

    def __init__(self, instance_of: Type[T] = None, default: T = None, **kwargs):
        self.instance_of = instance_of
        self.default = default
        self._initialized = self._initialized

    def __call__(self, *args, **kwargs) -> T:
        # if not self._initialized and not self.default:
        #     raise AttributeError("Dependency must me initialized with from_initialized() or with default kwarg")
        if not self._initialized:
            return self.default
        return self._initialized

    def from_initialized(self, obj) -> Dependency[T]:
        if not isinstance(obj, self.instance_of) or not issubclass(obj.__class__, self.instance_of):
            raise AttributeError("Dependency must init with same instance as instance_of or nested cls")
        self._initialized = obj
        return self

    @property
    def initialized(self) -> T:
        return self._initialized



from abc import ABCMeta
from copy import copy
from typing import Any

from orjson import dumps, loads

from web_foundation.config import GenericConfig
from web_foundation.resources.stores.events import StoreUpdateEvent
from web_foundation.kernel.messaging.channel import IChannel
from web_foundation.resources.resource import Resource
from web_foundation.utils.types import TypeJSON


class DataStore(Resource, metaclass=ABCMeta):
    need_sync: bool = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    async def on_store_update(self, event: StoreUpdateEvent):
        if not event.obj:
            event.value = self._load(event.value)
        await self.set_item(event.key, event.value, send_event=False)

    async def get_item(self, key: str) -> Any:
        raise NotImplementedError

    async def _set_item(self, key: str, value: TypeJSON):
        raise NotImplementedError

    async def set_item(self, key: str, value: Any, obj: bool = False, send_event: bool = True):
        await self._set_item(key, value)
        if self.need_sync and send_event:
            data = value if obj else self._serialize(value)
            await self._channel.produce(StoreUpdateEvent(key, copy(data), obj))

    async def get_all(self):
        raise NotImplementedError

    async def size(self):
        raise NotImplementedError

    async def init(self, config: GenericConfig, channel: IChannel, **kwargs):
        await super(DataStore, self).init(config, channel)
        self._channel.add_event_listener(StoreUpdateEvent, self.on_store_update)

    def _load(self, value: str):
        return loads(value)

    def _serialize(self, value: TypeJSON):
        return dumps(value)



from typing import List, Dict, Any, AnyStr

from sanic.response import raw

from web_foundation.extentions.metrics.metric import Metric, CounterMetric, SummaryMetric, \
    HistogramMetric
from web_foundation.utils.types import TypeJSON


class MetricExporter:
    @classmethod
    async def export(cls, metrics: List[Metric]) -> Dict[str, Any]:
        raise NotImplementedError

    @classmethod
    def empty(cls):
        return {}


class JsonExporter(MetricExporter):
    @classmethod
    async def export(cls, metrics: List[Metric]) -> Dict[str, TypeJSON]:
        dic = {}
        for metric in metrics:
            if not metric:
                continue
            if not dic.get(metric.name):
                dic[metric.name] = [{"kind": metric.kind,
                                     "value": metric.collect(),
                                     "labels": metric.labels,
                                     "timestamp": metric.timestamp}]
            else:
                dic[metric.name].append({"kind": metric.kind,
                                         "value": metric.collect(),
                                         "labels": metric.labels,
                                         "timestamp": metric.timestamp})
        return dic


class PrometheusExporter(MetricExporter):
    @classmethod
    async def export(cls, export_metrics: List[Metric]) -> AnyStr:
        try:
            from prometheus_client import metrics, registry
            from prometheus_client import metrics
            from prometheus_client.exposition import generate_latest
            from prometheus_client.metrics_core import CounterMetricFamily, \
                HistogramMetricFamily, \
                SummaryMetricFamily, \
                GaugeMetricFamily
            from prometheus_client.registry import Collector

            class CustomCollector(Collector):
                _custom_metrics: List[Metric]

                def __init__(self, custom_metrics: List[Metric]):
                    self._custom_metrics = custom_metrics

                def collect(self):
                    exp_metric = []
                    for metr in self._custom_metrics:
                        if isinstance(metr, CounterMetric):
                            prm_metr = CounterMetricFamily(metr.name, metr.description, labels=metr.labels_keys)
                            prm_metr.add_metric(metr.labels_values, metr.value, timestamp=metr.timestamp)
                            yield prm_metr
                        elif isinstance(metr, SummaryMetric):
                            prm_metr = SummaryMetricFamily(metr.name, metr.description, labels=metr.labels_keys)
                            prm_metr.add_metric(labels=metr.labels_values,
                                                sum_value=metr.value,
                                                count_value=metr.count,
                                                timestamp=metr.timestamp)
                            yield prm_metr
                        elif isinstance(metr, HistogramMetric):
                            prm_metr = HistogramMetricFamily(metr.name, metr.description, labels=metr.labels_keys)
                            prm_metr.add_metric(labels=metr.labels_values,
                                                sum_value=metr.sum,
                                                buckets=metr.buckets_list,
                                                timestamp=metr.timestamp)
                            yield prm_metr

            reg = registry.CollectorRegistry()
            reg.register(CustomCollector(export_metrics))
            return generate_latest(reg)
            # return  self._get_metrics_data()
        except ImportError:
            raise ImportError(
                "You need to install prometheus-client to use PrometheusExporter (web-foundation[prometheus])")

    @staticmethod
    def response_fabric(data):
        return raw(data, content_type='text/plain; version=0.0.4; charset=utf-8')



from __future__ import annotations

import asyncio
import time
from functools import partial
from types import SimpleNamespace
from typing import Type, Generic, Callable, Dict, Any, List

import orjson
from sanic import Sanic, Request
from sanic.response import BaseHTTPResponse

from web_foundation import settings
from web_foundation.config import AppConf
from web_foundation.extentions.error_handler import ExtendedErrorHandler
from web_foundation.extentions.metrics.events import NewMetricEvent
from web_foundation.extentions.metrics.metric import HistogramMetric, CounterMetric, SummaryMetric, Metric
from web_foundation.extentions.realtime.broadcaster import RtBroadcaster
from web_foundation.extentions.realtime.connection import RtEventCallback
from web_foundation.extentions.router import ExtRouter
from web_foundation.extentions.sanic_ext_worker import ExtWorker
from web_foundation.kernel.container import GenericDependencyContainer
from web_foundation.kernel.isolate import Isolate
from web_foundation.kernel.messaging.channel import IChannel, IMessage, GenericIMessage
from web_foundation.kernel.messaging.dispatcher import IDispatcher, IDispatcherIsolate
from web_foundation.kernel.stage import AppStage, StageListener
from web_foundation.settings_event import SettingsChangeEvent


class App(Generic[GenericDependencyContainer]):
    name: str
    container: GenericDependencyContainer
    sanic: Sanic
    config: AppConf
    dispatcher: IDispatcher
    _rt_listeners = list[tuple[list[Type[IMessage] | str], RtEventCallback, bool]]
    _stage_listeners: Dict[AppStage, List[StageListener]]

    def __init__(self, dependency_container: GenericDependencyContainer,
                 dispatcher: IDispatcher = None):
        self.config = dependency_container.app_config()
        self.name = dependency_container.app_config().app_name
        self.container = dependency_container
        self.sanic = Sanic(self.name, loads=orjson.loads, dumps=orjson.dumps,
                           error_handler=ExtendedErrorHandler())

        self.dispatcher = dispatcher if dispatcher else IDispatcher()
        self._rt_listeners = []
        self._stage_listeners = {}
        self.container.set_listeners(self.listeners_registrator)
        self.add_isolate("Dispatcher", IDispatcherIsolate, dispatcher=self.dispatcher, add_channel=False, daemon=True)

    def add_isolate(self, name: str, isolate_cls: Type[Isolate], daemon=True, add_channel: bool = True, **kwargs):
        """
        Add process to execute with worker manager
        """
        instance = isolate_cls(name, container=self.container, **kwargs)
        if add_channel:
            self.dispatcher.register_isolate(instance)

        def add_nondaemon(s_app: Sanic):
            nonlocal instance

            def override_manage(ident, func, kwargs: Dict[str, Any], transient=False, daemon=True):
                nonlocal s_app
                container = s_app.manager.transient if transient else s_app.manager.durable
                container.append(
                    ExtWorker(ident, func, kwargs, s_app.manager.context, s_app.manager.worker_state, daemon=daemon))

            s_app.manager.manage = override_manage
            s_app.manager.manage(name, instance.run_forked, {}, daemon=daemon)

        if daemon:
            self.sanic.main_process_ready(
                lambda s_app: s_app.manager.manage(name, instance.run_forked, {}))
            # self.sanic.main_process_ready(lambda s_app:s_app.manager.)
        else:
            self.sanic.main_process_ready(lambda s_app: add_nondaemon(s_app))

    def subscribe_worker(self, event_type: list[GenericIMessage | str] | GenericIMessage | str,
                         resolve_callback: RtEventCallback, use_nested_classes: bool = False):
        """
        Subscribe all sanic workers to event[s]
        :param use_nested_classes:
        :param event_type: Name or Class event
        :param resolve_callback: Resolve function
        """
        self._rt_listeners.append((event_type, resolve_callback, use_nested_classes))

    def listeners_registrator(self, stage: AppStage, callback: StageListener):
        if not self._stage_listeners.get(stage):
            self._stage_listeners[stage] = []
        self._stage_listeners[stage].append(callback)

    def add_custom_router(self, router: ExtRouter):
        router.ctx.app = self.sanic
        self.sanic.router = router

    async def _dispatch_stage(self, stage: AppStage):
        if not stage in self._stage_listeners:
            return
        for listener in self._stage_listeners[stage]:
            await listener(stage, self.container)

    def _set_sanic_confs(self):
        self.sanic.config.SWAGGER_UI_CONFIGURATION = {
            "docExpansion": 'none'
        }

    def _configure_sanic_main(self):
        async def on_start(s_app: Sanic):
            for i in range(s_app.state.workers):
                channel = IChannel(f"Channel_{i}")
                setattr(s_app.shared_ctx, f"Channel_{i}", channel)
                self.dispatcher.channels.update({f"Channel_{i}": channel})
                # noinspection PyAsyncCall
                # self.sanic.add_task(channel.listen_produce(self.dispatcher.on_channel_sent))

        self.sanic.main_process_ready(on_start)

    def _configure_sanic_worker(self):
        async def set_worker_ctx(s_app: Sanic):
            # set channel to ctx
            worker_num = int(s_app.m.name.split('-')[-2])
            s_app.ctx.channel = getattr(s_app.shared_ctx, f"Channel_{worker_num}")
            s_app.ctx.container = self.container
            for service, name in s_app.ctx.container.services():
                service.channel = s_app.ctx.channel
            s_app.ctx.rt_broadcaster = RtBroadcaster()
            for event, resolver, use_nested_classes in self._rt_listeners:
                s_app.ctx.channel.add_event_listener(
                    event_type=event,
                    callback=partial(s_app.ctx.rt_broadcaster.broadcast, resolve_callback=resolver),
                    use_nested_classes=use_nested_classes)
            await self.container.init_resources(s_app.ctx.channel)
            if loader := self.container.addon_loader():
                await loader.first_init()

        async def reg_metrics(s_app: Sanic):
            async def give_metric(metric: Metric):
                asyncio.create_task(s_app.ctx.channel.produce(NewMetricEvent(metric)))

            self.sanic.ctx.metrics = SimpleNamespace()

            async def metrics_before(request: Request):
                if not settings.METRICS_ENABLE:
                    return
                request.app.ctx.metrics.exec_time = time.time()

            async def metrics_after(request: Request, response: BaseHTTPResponse):
                if not settings.METRICS_ENABLE:
                    return
                else:
                    metr = HistogramMetric(f"exec_request_time")
                    metr.observe(time.time() - request.app.ctx.metrics.exec_time)
                    metr.add_label(request=request.path,
                                   method=request.method,
                                   status=str(response.status))
                    await give_metric(metr)
                    metr = CounterMetric(f"request_path_count", value=1)
                    metr.add_label(request=request.path,
                                   method=request.method,
                                   status=str(response.status))
                    await give_metric(metr)
                    metr = SummaryMetric(f"request_path_latency")
                    metr.add_label(request=request.path,
                                   method=request.method,
                                   status=str(response.status))
                    metr.observe(time.time() - request.app.ctx.metrics.exec_time)
                    await give_metric(metr)

            self.sanic.register_middleware(metrics_before)
            self.sanic.register_middleware(metrics_after, "response")

        self.sanic.before_server_start(set_worker_ctx)
        self.sanic.before_server_start(reg_metrics)

        async def listen_channel(s_app: Sanic):
            async def on_settings_change(event: SettingsChangeEvent):
                setattr(settings, event.name, event.value)

            s_app.ctx.channel.add_event_listener(SettingsChangeEvent, on_settings_change)
            # noinspection PyAsyncCall
            s_app.add_task(s_app.ctx.channel.listen_consume())

        self.sanic.after_server_start(listen_channel)

        async def on_stop(s_app: Sanic):
            await self.container.shutdown_resources()

        self.sanic.before_server_stop(on_stop)

    def run(self):
        asyncio.run(self._dispatch_stage(AppStage.BEFORE_APP_RUN))
        self.sanic.router.parse()
        self.sanic.router.apply_routes(self.sanic, self.container)
        self._set_sanic_confs()
        self._configure_sanic_main()
        self._configure_sanic_worker()
        asyncio.run(self._dispatch_stage(AppStage.BEFORE_SERVER_START))
        sanic_conf = self.container.app_config().sanic
        start_args = sanic_conf.dict() if sanic_conf else {}
        self.sanic.run(**start_args)


AppListener = Callable[[App], None]



import asyncio
from asyncio import Future, CancelledError
from typing import Dict, Any, List

import loguru

from web_foundation import settings
from web_foundation.extentions.metrics.metric import Metric, CounterMetric
from web_foundation.extentions.metrics.events import NewMetricEvent, MetricRequestEvent, MetricResponseEvent
from web_foundation.kernel.messaging.channel import IChannel, EventListener, GenericIMessage
from web_foundation.kernel.isolate import Isolate
from web_foundation.extentions.schedulle.scheduller import TaskScheduler, TaskIMessage
from web_foundation.settings_event import SettingsChangeEvent

MetricsDict = Dict[str, Dict[str, Metric]]


class IDispatcher:
    channels: Dict[str, IChannel]
    events_listeners: Dict[str, EventListener]
    scheduler: TaskScheduler
    collected_metrics: Dict[str, MetricsDict]
    _msg_global_index: int
    state: Dict[str, Any]

    def __init__(self, scheduler: TaskScheduler = None):
        self.scheduler = scheduler
        self.state = {}
        self.channels = {}
        self.events_listeners = {}
        self.collected_metrics = {}
        self.collected_metrics.update({"__system_metric__": {}})
        self._msg_global_index = 0
        if self.scheduler:
            self.scheduler.produce_method = self.broadcast
            self.state.update({"scheduler": {}})
            self.scheduler.state = self.state.get("scheduler")
            self.add_event_listener(TaskIMessage.message_type, self.scheduler.add_task)
        if settings.METRICS_ENABLE:
            self.add_event_listener(NewMetricEvent.message_type, self.on_new_metric)
            self.add_event_listener(MetricRequestEvent.message_type, self.on_metric_request)

        async def on_settings_change(event: SettingsChangeEvent):
            setattr(settings, event.name, event.value)

        self.add_event_listener(SettingsChangeEvent.message_type, on_settings_change)

    def register_isolate(self, isolate: Isolate):
        self.channels.update({isolate.name: isolate.channel})

    def add_event_listener(self, event_type: str, callback: EventListener):
        self.events_listeners[event_type] = callback

    async def track_event(self, msg: GenericIMessage):
        listener = self.events_listeners.get(msg.message_type)
        if listener:
            await listener(msg)

        if settings.METRICS_ENABLE:
            event_counter = CounterMetric("events_counter", self._msg_global_index)
            save_to = self.collected_metrics['__system_metric__']
            save_to.update({"events_counter": {"dispatcher_events_counter": event_counter}})

            named_event_counter = save_to.get("named_events_counter")
            if not named_event_counter:
                mtr = CounterMetric("named_events_counter", value=1)
                mtr.add_label(event_name=msg.message_type)
                named_event_counter = {msg.message_type: mtr}
                save_to["named_events_counter"] = named_event_counter
            elif named_event_counter.get(msg.message_type):
                named_event_counter[msg.message_type].inc()
            else:
                mtr = CounterMetric("named_events_counter", value=1)
                mtr.add_label(event_name=msg.message_type)
                named_event_counter[msg.message_type] = mtr
            if settings.DEBUG:
                loguru.logger.debug(f'MetricsDispatcher - track event {msg}')

    async def on_channel_sent(self, msg: GenericIMessage):
        self._msg_global_index += 1
        msg.index = self._msg_global_index
        asyncio.create_task(self.track_event(msg))
        if msg.destination != "__dispatcher__":
            asyncio.create_task(self.broadcast(msg))

    async def on_metric_request(self, event: MetricRequestEvent):
        exporter = event.exporter
        sender_channel = self.channels.get(event.sender)
        if sender_channel:
            if settings.DEBUG:
                loguru.logger.debug(f"Send METRICS to {sender_channel}")
            if not self.collected_metrics.get(event.kind) or not self.collected_metrics[
                event.kind] or not settings.METRICS_ENABLE:
                resp = MetricResponseEvent(exporter.empty())
                resp.inner_index = event.inner_index
                await sender_channel.sent_to_consume(resp)
                return
            rp = []
            for i in self.collected_metrics[event.kind].values():
                if isinstance(i, dict):
                    for k in i.values():
                        rp.append(k)
                else:
                    rp.append(i)
            data = await exporter.export(rp)
            resp = MetricResponseEvent(data)
            resp.inner_index = event.inner_index
            await sender_channel.sent_to_consume(resp)

    async def on_new_metric(self, event: NewMetricEvent):
        if not self.collected_metrics.get(event.metric.kind):
            self.collected_metrics.update({event.metric.kind: {}})
        metr = self.collected_metrics[event.metric.kind].get(event.metric.name)
        if not metr:
            self.collected_metrics[event.metric.kind].update(
                {event.metric.name: {event.metric.labels_concat: event.metric}})
        else:
            if metr.get(event.metric.labels_concat):
                # loguru.logger.warning(f"get {event.metric}")
                metr[event.metric.labels_concat] += event.metric
            else:
                # loguru.logger.warning(f"set {event.metric}")
                metr[event.metric.labels_concat] = event.metric

    async def broadcast(self, msg: GenericIMessage):
        for worker_name, ch in self.channels.items():
            if msg.sender == worker_name:
                continue
            if msg.destination in ["__all__", worker_name]:
                if settings.DEBUG:
                    loguru.logger.debug(f"Sent {msg} to {ch.worker_name}")
                ch.consume_pipe.put(msg)

    def perform(self, **kwargs) -> List[Future[Any]]:
        tasks = []
        for channel in self.channels.values():
            tasks.append(asyncio.ensure_future(
                channel.listen_produce(
                    self.on_channel_sent
                )))
        return tasks


class IDispatcherIsolate(Isolate):
    dispatcher: IDispatcher

    def __init__(self, name: str, dispatcher: IDispatcher, **kwargs):
        self.dispatcher = dispatcher
        super().__init__(name, **kwargs)

    async def work(self, **kwargs) -> None:
        if self.dispatcher.scheduler:
            self.dispatcher.scheduler.perform()
        try:
            await asyncio.wait(self.dispatcher.perform())
        except CancelledError:
            pass


from abc import abstractmethod, ABCMeta
from typing import Any, Dict, List

from web_foundation.logic.module import AbstractModule
from pydantic import BaseModel


class AbstractUseCase(metaclass=ABCMeta):
    _addictions: Dict[str, AbstractModule]
    name: str | None

    version: int
    method: str
    route: str
    in_model: BaseModel
    out_model: BaseModel

    roles: List[str]

    def raise_if_none(self, val: Any, ex: Exception = Exception("Item not found")):
        if not val:
            raise ex

    @abstractmethod
    async def execute(self, *args, **kwargs):
        pass

    @property
    def available(self):
        return all([i.available for i in self._addictions.values()])



from typing import Any

from web_foundation.kernel.messaging.channel import IMessage


class StoreUpdateEvent(IMessage):
    message_type = "store_update"
    obj: bool = False

    def __init__(self, key: str, value: Any, obj: bool = False):
        super().__init__()
        self.key = key
        self.value = value
        self.obj = obj

    def __str__(self):
        return f"{self.__class__.__name__}({self.key})"



from typing import Any

from web_foundation.extentions.metrics.metric import Metric
from web_foundation.extentions.metrics.exporter import MetricExporter
from web_foundation.kernel.messaging.channel import IMessage


class NewMetricEvent(IMessage):
    message_type = "new_metric"
    destination = "__dispatcher__"
    metric: Metric

    def __init__(self, metric: Metric):
        super().__init__()
        self.metric = metric

    def __str__(self):
        return f"NewMetricEvent({self.metric.__str__()})"


class MetricRequestEvent(IMessage):
    message_type = "metrics_request"
    destination = "__dispatcher__"
    exporter: MetricExporter
    kind: str

    def __init__(self, exporter: MetricExporter, kind: str):
        super().__init__()
        self.kind = kind
        self.exporter = exporter


class MetricResponseEvent(IMessage):
    message_type = "metrics_response"
    metrics_data: Any

    def __init__(self, metrics_data: Any):
        super().__init__()
        self.metrics_data = metrics_data

    def __str__(self):
        return f"MetricResponse({type(self.metrics_data)})"



from __future__ import annotations

from multiprocessing.context import BaseContext
from typing import Dict, Any
from sanic.worker.manager import Worker, WorkerProcess
from sanic.worker.process import ProcessState


class ExtWorkerProcess(WorkerProcess):
    daemon: bool

    def __init__(self, factory, name, target, kwargs, worker_state, daemon=True):
        self.daemon = daemon
        super().__init__(factory, name, target, kwargs, worker_state)

    def spawn(self):
        if self.state is not ProcessState.IDLE:
            raise Exception("Cannot spawn a worker process until it is idle.")
        self._process = self.factory(
            name=self.name,
            target=self.target,
            kwargs=self.kwargs,
            daemon=self.daemon,
        )


class ExtWorker(Worker):
    def __init__(self, ident: str, serve, server_settings, context: BaseContext, worker_state: Dict[str, Any],
                 daemon=True):
        self.daemon = daemon
        super().__init__(ident, serve, server_settings, context, worker_state)

    def create_process(self) -> WorkerProcess:
        process = ExtWorkerProcess(
            factory=self.context.Process,
            name=f"Sanic-{self.ident}-{len(self.processes)}",
            target=self.serve,
            kwargs={**self.server_settings},
            worker_state=self.worker_state,
            daemon=self.daemon
        )
        self.processes.add(process)
        return process



from enum import Enum
from typing import Callable, Coroutine, Any


class AppStage(Enum):
    BEFORE_APP_RUN = 0
    BEFORE_SERVER_START = 1


StageListener = Callable[[AppStage, 'GenericDependencyContainer'], Coroutine[Any, Any, None]]
ListenerRegistrator = Callable[[AppStage, StageListener], None]



import asyncio
import signal
from abc import ABCMeta
from asyncio import CancelledError
from typing import Dict, Generic

import loguru
from sanic import Sanic
from sanic.worker.multiplexer import WorkerMultiplexer

from web_foundation.extentions.sanic_ext_worker import ExtWorker
from web_foundation.kernel.container import GenericDependencyContainer
from web_foundation.kernel.messaging.channel import IChannel


class Isolate(Generic[GenericDependencyContainer], metaclass=ABCMeta):
    name: str
    channel: IChannel | None
    container: GenericDependencyContainer
    init_kwargs: Dict

    def __init__(self, name: str, container: GenericDependencyContainer, **kwargs):
        self.name = name
        self.container = container
        self.channel = IChannel(f"Channel_{name}")
        self.init_kwargs = kwargs

    async def work(self, **kwargs) -> None:
        raise NotImplementedError

    def run_forked(self) -> None:
        async def call():
            try:
                nonlocal self
                await self.container.init_resources(self.channel)
                asyncio.create_task(self.channel.listen_consume())
                await self.work(**self.init_kwargs)
                await self.container.shutdown_resources()
            except CancelledError:
                await self.container.shutdown_resources()
                self.close()

        def signal_handler(sig, frame):
            for task in asyncio.all_tasks():
                task.cancel()

        signal.signal(signal.SIGINT, signal_handler)

        try:
            asyncio.run(call())
        except KeyboardInterrupt:
            asyncio.run(self.container.shutdown_resources())
            self.close()

    def close(self):
        pass



from typing import Any

from web_foundation.config import AppConf
from web_foundation.kernel.messaging.channel import IChannel
from web_foundation.kernel.stage import ListenerRegistrator


class Resource:
    _communicator: Any | None
    _channel: IChannel | None
    _inited: bool
    _config: AppConf

    def __init__(self, *args, **kwargs):
        _inited = False

    async def init(self, config: AppConf, channel: IChannel | None = None, communicator: Any | None = None, *args,
                   **kwargs) -> None:
        self._config = config
        self._channel = channel
        self._inited = True
        self._communicator = communicator

    async def shutdown(self):
        raise NotImplementedError

    def clear(self):
        self._inited = False
        self._communicator = None
        self._channel = None

    def set_listeners(self, registrator: ListenerRegistrator):
        pass

    @property
    def ready(self):
        return self._inited

    @property
    def channel(self):
        if not self._inited:
            raise RuntimeError("Resource not inited")
        return self._channel

    @property
    def communicator(self):
        if not self._inited:
            raise RuntimeError("Resource not inited. Can't access to communicator")
        return self._communicator



import inspect
import re
from copy import deepcopy, copy
from functools import wraps
from typing import Callable, Type

import sanic_ext
from sanic_ext.extensions.openapi.types import Schema, Array
from tortoise import Model


def set_access_endpoints(access_endpoints: dict, models,
                         db_model_responses: dict[Type[Model]], get_out_struct_func: Callable,
                         get_list_out_struct_func: Callable, drop_filter_fields: list[str],
                         scope: str = None):
    params_filters = {
        "not": Schema(description="not definition", type="string"),

        "in": Array(Schema(description="any field value", type="string"),
                    description="checks if value of field is in passed list"),

        "not_in": Array(Schema(description="any field value", type="string"),
                        description="not in definition"),

        "gte": Schema(description="greater or equals than passed value", type="string"),

        "gt": Schema(description="greater than passed value", type="string"),

        "lte": Schema(description="lower or equals than passed value", type="string"),

        "lt": Schema(description="lower than passed value", type="string"),

        "range": Array(Schema(description="any field value", type="string"),
                       description="between and given two values",
                       maxItems=2, minItems=2),

        "isnull": Schema(description="field is null", type="boolean"),

        "not_isnull": Schema(description="field is not null", type="boolean"),

        "contains": Schema(description="field contains specified substring", type="string"),

        "icontains": Schema(description="case insensitive contains", type="string"),

        "startswith": Schema(description="if field starts with value", type="string"),

        "istartswith": Schema(description="case insensitive startswith", type="string"),

        "endswith": Schema(description="if field ends with value", type="string"),

        "iendswith": Schema(description="case insensitive endswith", type="string"),

        "iexact": Schema(description="case insensitive equals", type="string"),
    }

    for endpoint, versions in deepcopy(access_endpoints).items():
        for version, params in deepcopy(versions).items():
            endpoint_handler = params.get("handler")
            endpoint_protector = params.get("protector")
            endpoint_response_fabric = params.get("response_fabric")
            if scope:
                access_endpoints[endpoint][version].update({'scope': scope})
            for method_name, param in deepcopy(params).items():
                if not isinstance(param, dict):
                    continue
                target_func = param.get('handler')
                target_func = target_func if target_func else endpoint_handler
                method_protector = param.get("protector")
                method_protector = method_protector if method_protector else endpoint_protector
                if param.get('out_dto'):
                    out_struct = param.get('out_dto')
                else:
                    # get out_struct from model name (in exec_access args)
                    func_text = inspect.getsource(target_func)
                    model_name = re.findall(r"context, (\w*)", func_text)
                    if model_name:
                        model = getattr(models, model_name[0])
                        out_struct = db_model_responses.get(model)
                        if not out_struct:
                            out_struct = get_out_struct_func(model)
                        access_endpoints[endpoint][version][method_name]['out_dto'] = out_struct
                        param['out_dto'] = out_struct
                    else:
                        out_struct = None

                # add routes with <entity_id>
                if method_name in ['patch', 'delete', 'get']:
                    if method_name != 'get':
                        access_endpoints[endpoint][version].pop(method_name)
                    if not access_endpoints.get(f"{endpoint}/<entity_id:int>"):
                        access_endpoints.update({f"{endpoint}/<entity_id:int>": {}})

                    if not access_endpoints[f"{endpoint}/<entity_id:int>"].get(version):
                        access_endpoints[f"{endpoint}/<entity_id:int>"].update({
                            version: {
                                'handler': endpoint_handler,
                                'protector': endpoint_protector,
                                'response_fabric': endpoint_response_fabric
                            }
                        })
                    access_endpoints[f"{endpoint}/<entity_id:int>"][version].update({method_name: copy(param)})
                    if scope:
                        access_endpoints[f"{endpoint}/<entity_id:int>"][version]['scope'] = scope

                    if method_name == 'get' and out_struct:
                        access_endpoints[endpoint][version][method_name]['out_dto'] = get_list_out_struct_func(
                            out_struct)
                    if not access_endpoints[endpoint].get(version):
                        access_endpoints[endpoint].pop(version)

                def new_wrap(func):  # wrapper to change handler signature (need for openapi)
                    @wraps(func)
                    def wr(*args, **kwargs):
                        return func(*args, **kwargs)

                    return wr

                if method_name == 'get' and 'entity_id' not in endpoint and out_struct:
                    get_access_func = new_wrap(target_func)
                    get_access_func.__name__ = f"{get_access_func.__name__}__read"
                    for field_model in out_struct.__fields__.values():
                        if field_model.name in drop_filter_fields:
                            continue
                        for filter, schema in params_filters.items():
                            schema.type = str
                            schema.example = f"{field_model.name}__{filter}"
                            get_access_func = sanic_ext.openapi.parameter(
                                name=f"{field_model.name}__{filter}", schema=schema)(get_access_func)
                    access_endpoints[endpoint][version][method_name]['handler'] = get_access_func


def set_security_scheme(routes_dict: dict, protectors_dict: dict[Callable, dict[str, dict | list]]):
    """
    protectors_dict: {protector: {"scheme": dict, Optional("additional_parameters": list[Decorator])}
        example: {protector: {
                    "scheme": {"cookieAuth": ["user"]},
                    "additional_parameters": [sanic_ext.openapi.parameter(
                                name=f"session", schema=int, location="cookie", required=True)]
    """
    for app_route in routes_dict.get("apps"):
        secured_handlers = []
        for endpoint, versions in app_route.get("endpoints").items():
            for version, params in versions.items():
                endpoint_handler = params.get("handler", None)
                endpoint_protector = params.get("protector", None)
                for method_name, method_params in params.items():
                    if not isinstance(method_params, dict):
                        continue
                    target_func = method_params.get('handler')
                    target_func = target_func if target_func else endpoint_handler
                    protector = method_params.get("protector")
                    protector = protector if protector else endpoint_protector
                    for _protector, protector_params in protectors_dict.items():
                        if protector == _protector:
                            if secure_params := protector_params.get("additional_parameters"):
                                if target_func not in secured_handlers:
                                    for secure_param in secure_params:
                                        secure_param(target_func)
                                secured_handlers.append(target_func)
                            if scheme := protector_params.get("scheme"):
                                sanic_ext.openapi.secured(scheme)(target_func)



from __future__ import annotations

from datetime import datetime
from enum import Enum
from typing import List, Sequence, Dict, Tuple


class Metric:
    name: str
    value: float
    kind: str
    labels: Dict[str, str]
    description: str
    timestamp: float

    def __init__(self, name: str, value: float = 0.0, labels: Dict[str, str] = None, description: str = None,
                 kind: str = "__system_metric__"):
        self.kind = kind
        self.labels = labels if labels else {}
        self.timestamp = datetime.now().timestamp()
        self.description = description if description else 'Default metric description'
        self.name = name
        self.value = float(value)

    def add_label(self, name: str = None, value: str = None, **kwargs):
        if kwargs:
            self.labels.update(kwargs)
        else:
            self.labels.update({name: value})

    def set(self, value: float):
        self.value = value

    @property
    def labels_keys(self):
        return list(self.labels.keys())

    @property
    def labels_values(self):
        return list(self.labels.values())

    @property
    def labels_concat(self):
        return '_'.join(self.labels.values())

    def inc(self, value: int = 1):
        raise NotImplementedError

    def __str__(self):
        return f"{self.__class__.__name__}({self.name})<{self.value}>"

    def __repr__(self):
        return f"{self.__class__.__name__}({self.name})<{self.value}>"

    def __eq__(self, other):
        return self.name == other.name

    def __add__(self, metric: Metric):
        self.value += metric.value
        return self

    def __hash__(self):
        return id(self.name)

    def collect(self):
        return self.value


class CounterMetric(Metric):
    def inc(self, value: int = 1):
        self.value += value


class SummaryMetric(Metric):
    _count: int = 0

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._count = 1

    def __add__(self, other: SummaryMetric):
        self._count += other._count
        self.value += other.value
        return self

    def observe(self, value: int = 1):
        self._count += 1
        self.value += value

    def collect(self):
        return self.value / self._count

    def __str__(self):
        return f"{self.__class__.__name__}({self.name})<{self.collect()}>"

    def __repr__(self):
        return f"{self.__class__.__name__}({self.name})<{self.collect()}>"

    def inc(self, value: int = 1):
        self.value += value

    @property
    def count(self):
        return self._count


class HistogramMetric(Metric):
    _sum: float = 0.0
    _buckets: List[float]
    _temp_bucket: Sequence[float]
    DEFAULT_BUCKETS = (.001, .005, .01, .025, .05, .075, .1, .25, .5, .75, 1.0, 2.5, 5.0, 7.5, 10.0, float("inf"))

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._sum = 0.0
        self._temp_bucket = self.DEFAULT_BUCKETS
        self._buckets = [0.0 for _ in range(len(self._temp_bucket))]

    def set_bucket(self, buckets: Sequence[float]):
        buckets = [float(b) for b in buckets]
        if buckets != sorted(buckets):
            raise ValueError('Buckets not in sorted order')
        if buckets and buckets[-1] != float("inf"):
            buckets.append(float("inf"))
        if len(buckets) < 2:
            raise ValueError('Must have at least two buckets')
        self._temp_bucket = buckets
        self._buckets = [0.0 for _ in range(len(self._temp_bucket))]

    def __add__(self, other: HistogramMetric):
        if self._temp_bucket != other._temp_bucket:
            raise ValueError("Buckets of two Histogram is not comparable")
        self._sum += other._sum
        for i, bound in enumerate(self._temp_bucket):
            self._buckets[i] += other._buckets[i]
        return self

    def observe(self, value: float):
        self._sum += value
        for i, bound in enumerate(self._temp_bucket):
            if value <= bound:
                self._buckets[i] += 1

    def collect(self):
        return list(zip(self._temp_bucket, self._buckets))

    def __str__(self):
        return f"{self.__class__.__name__}({self.name})<{self._sum}>"

    def __repr__(self):
        return f"{self.__class__.__name__}({self.name})<{self._sum}>"

    def inc(self, value: int = 1):
        self.value += value

    @property
    def sum(self):
        return self._sum

    @property
    def buckets_list(self) -> List[Tuple[str, float]]:
        lst = []
        for i in range(len(self._buckets)):
            lst.append((str(self._temp_bucket[i]), self._buckets[i]))
        return lst



from __future__ import annotations

import asyncio
from multiprocessing import Queue
from typing import Coroutine, Callable, Dict, Type, List, Any
from typing import TypeVar

from loguru import logger

from web_foundation import settings


class IMessage:
    message_type: str = "__all__"
    index: int
    inner_index: int
    sender: str
    destination: str = "__all__"
    exec_inner: bool = False

    def __init__(self):
        self.index = 0
        self.inner_index = 0
        self.sender = "None"

    def __str__(self):
        return f"{self.__class__.__name__}({self.message_type},{self.index=},{self.inner_index=}, {self.sender=})"


GenericIMessage = TypeVar("GenericIMessage", bound=IMessage, contravariant=True)

EventListener = Callable[[GenericIMessage], Coroutine]


class IMessageWaiter:
    event = asyncio.Event()
    result: Any = None
    call = None

    def __eq__(self, other):
        self.wait_for_index = other.wait_for_index


class IChannel:
    read_timeout = 0.01
    worker_name: str
    consume_pipe: Queue
    produce_pipe: Queue

    _listeners: Dict[str, List[EventListener]]
    _response_waiters: Dict[int, IMessageWaiter]

    _before_event_middlewares: Dict[str, EventListener]
    _after_event_middlewares: Dict[str, EventListener]
    _inner_index: int

    def __init__(self, isolate_name: str):
        self._inner_index = 0
        self.worker_name = isolate_name
        # self.debug = debug
        self._listeners = {}
        self._response_waiters = {}
        self.consume_pipe = Queue()
        self.produce_pipe = Queue()
        self._after_event_middlewares = {}
        self._before_event_middlewares = {}

    def __str__(self):
        return f"IChannel(name={self.worker_name})"

    async def produce(self, msg: IMessage):
        self._inner_index += 1
        msg.inner_index = self._inner_index
        msg.sender = self.worker_name
        before_middleware = self._before_event_middlewares.get(msg.message_type)
        if before_middleware:
            await before_middleware(msg)
        if msg.exec_inner:
            if listeners := self._listeners.get(msg.message_type):
                for listener in listeners:
                    asyncio.create_task(listener(msg))
        self.produce_pipe.put(msg)

    async def sent_to_consume(self, msg: IMessage):
        self.consume_pipe.put(msg)

    async def listen_produce(self, callback: Callable[[GenericIMessage], Coroutine]):
        while True:
            while self.produce_pipe.empty():
                await asyncio.sleep(0.1)
            r: GenericIMessage = self.produce_pipe.get()
            if settings.DEBUG:
                logger.info(
                    f"Channel {self.worker_name} send message: {r.message_type}")
            await callback(r)

    async def listen_consume(self):
        while True:
            while self.consume_pipe.empty():
                await asyncio.sleep(0.1)
            r: IMessage = self.consume_pipe.get()
            if settings.DEBUG:
                logger.info(
                    f"Channel {self.worker_name} receive message: {r.message_type}")
            after_middleware = self._after_event_middlewares.get(r.message_type)
            if after_middleware:
                asyncio.create_task(after_middleware(r))
            response_waiter = self._response_waiters.get(r.inner_index)
            if response_waiter:
                response_waiter.result = r
            callbacks = self._listeners.get(r.message_type)
            if not callbacks:
                continue
            for callback in callbacks:
                asyncio.create_task(callback(r))

    def add_event_middleware(self, event_type: str, callback: EventListener, assign_to: str = "after"):
        """
        :param callback: callback to call
        :param event_type: "event_message_type"
        :param assign_to: "before, after"
        :return: None
        """
        match assign_to:
            case 'before':
                self._before_event_middlewares.update({event_type: callback})
            case "after":
                self._after_event_middlewares.update({event_type: callback})

    async def produce_for_response(self, msg: GenericIMessage) -> Any:
        task = IMessageWaiter()
        await self.produce(msg)
        self._response_waiters.update({msg.inner_index: task})
        while task.result is None:
            await asyncio.sleep(0.01)
        self._response_waiters.pop(msg.inner_index)
        return task.result

    def _add_event_listener(self, event_type: Type[IMessage] | str, callback: EventListener,
                            use_nested_classes: bool = False):
        if use_nested_classes and isinstance(event_type, str):
            raise AttributeError("Can't add_event_listener with use_nested_classes")

        def _add(event_name: str):
            nonlocal self
            nonlocal callback
            if event_name not in self._listeners:
                self._listeners[event_name] = []
            self._listeners[event_name].append(callback)

        if isinstance(event_type, str):
            _add(event_type)
        else:
            def _raise(cls_type):
                if not hasattr(cls_type, "message_type"):
                    raise AttributeError("Can't register listener cause message_type not found")

            if use_nested_classes:
                for cls in event_type.__subclasses__():
                    _raise(cls)
                    _add(cls.message_type)
            else:
                _raise(event_type)
                _add(event_type.message_type)

    def add_event_listener(self, event_type: List[Type[IMessage] | str] | Type[IMessage] | str,
                           callback: EventListener, use_nested_classes: bool = False):
        if isinstance(event_type, list):
            for i in event_type:
                self._add_event_listener(i, callback, use_nested_classes)
        else:
            self._add_event_listener(event_type, callback, use_nested_classes)

    def remove_event_listeners(self, event_type: Type[IMessage] | str):
        if isinstance(event_type, str):
            self._listeners.pop(event_type)
        else:
            self._listeners.pop(event_type.message_type)

    @property
    def listeners(self):
        return self._listeners



from typing import Any, Type
from typing import Union, Dict, List

JSON = Union[Dict[str, Any], List[Any], int, str, float, bool, Type[None]]
TypeJSON = Union[Dict[str, 'JSON'], List['JSON'], int, str, float, bool, Type[None]]



import asyncio
import json
import re
from datetime import datetime
from datetime import time as DTime
from functools import wraps, partial
from inspect import isawaitable, iscoroutine, iscoroutinefunction
from pathlib import Path
from typing import Union, List, Type, Callable, Coroutine, Any

from web_foundation.config import GenericConfig
from web_foundation.errors.app.application import InconsistencyError


# ------------------------------------------------STR_TO---------------------------------------------------------------
from web_foundation.kernel.dependency import Dependency


def str_to_list(string: str) -> Union[List[str], None]:
    if not ",":
        return None
    return string.split(",")


def str_to_bool(string: str) -> Union[bool, None]:
    try:
        if string.lower() == "false":
            return False
        elif string.lower() == "true":
            return True
        else:
            return None
    except Exception as e:
        return None


# ------------------------------------------------VALIDATE--------------------------------------------------------------


def validate_date(date: str, check_gt_now=True) -> str:
    try:
        if "." in date:
            _date = datetime.strptime(date, "%d.%m.%Y")
            date = _date.isoformat()
        else:
            _date = datetime.fromisoformat(date)
    except ValueError as exception:
        raise InconsistencyError(exception)
    if check_gt_now and datetime.now() > _date:
        raise InconsistencyError(message="Incorrect date")
    return date


def validate_time(str_time: str) -> str:
    try:
        t = DTime.fromisoformat(str_time)
    except ValueError as exception:
        raise InconsistencyError(exception)
    return str_time


def validate_datetime(date_time: str, check_gt_now=False) -> str:  # TODO add timezones
    format_types = ["%d", "%m", "%Y", "%H", "%M", "%S"]
    try:
        if "." in date_time and date_time.count(".") > 1:
            date_nums = [ddd for ddd in re.findall(r"(\d*)", date_time) if ddd]
            _date = datetime.strptime(' '.join(date_nums), ' '.join(format_types[:len(date_nums)]))
        else:
            _date = datetime.fromisoformat(date_time)
        date = _date.astimezone().isoformat()
    except ValueError as exception:
        raise InconsistencyError(exception)
    if check_gt_now and datetime.now() > _date:
        raise InconsistencyError(message="Incorrect datetime")
    return date


# ------------------------------------ OS ----------------------------------------

def load_config(conf_path: Path, config_model: Type[GenericConfig]) -> GenericConfig:
    """
    Load environment config to user in
    :param conf_path:
    :param config_model: BaseModel to cast json file to pydantic
    :return: None
    """
    with open(conf_path, "r") as _json_file:
        conf = config_model(**json.loads(_json_file.read()))
        return conf


# ----------------------------------------------SOME-------------------------------

def all_subclasses(cls):
    return set(cls.__subclasses__()).union(
        [s for c in cls.__subclasses__() for s in all_subclasses(c)])


def in_obj_subclasses(obj: object | Type, target_cls: Type):
    try:
        for i in dir(obj):
            if not i.startswith("__"):
                item = getattr(obj, i)
                if item.__class__ in all_subclasses(target_cls):
                    yield item, i
    except Exception as e:
        raise e


def in_obj_subcls_of_dependencies(obj: object | Type, target_cls: Type):
    try:
        for i in dir(obj):
            if not i.startswith("__"):
                item = getattr(obj, i)
                if isinstance(item, Dependency):
                    item = item.initialized
                if item.__class__ in all_subclasses(target_cls):
                    yield item, i
    except Exception as e:
        raise e


def in_obj_classes(obj: object | Type, target_cls: Type):
    try:
        for i in dir(obj):
            item = getattr(obj, i)
            if item.__class__ == target_cls:
                yield item, i
    except Exception as e:
        raise e


def in_obj_cls_items(obj: object | Type, target_cls: Type):
    try:
        for i in dir(obj):
            print()
            item = getattr(obj, i)
            cls = getattr(obj, i).__class__
            if item.__class__ == target_cls or cls in all_subclasses(target_cls):
                yield item, i
    except Exception as e:
        raise e


def get_callable_from_fnc(fnc: Callable[..., Coroutine[Any, Any, Any]] | partial):
    caller = fnc
    if iscoroutine(fnc) or isawaitable(fnc):
        @wraps(fnc)
        def __call(*args, **kwargs):
            asyncio.run(fnc(*args, **kwargs))

        caller = __call

    elif iscoroutinefunction(fnc):
        @wraps(fnc)
        def __call(*args, **kwargs):
            asyncio.run(fnc(*args, **kwargs))

        caller = __call

    return caller



from typing import Any, Dict, Mapping

from pydantic import BaseModel
from tortoise import Model


class CDto(BaseModel, Mapping):
    _dd: Dict[str, Any] = {}

    def __len__(self):
        return len(self.dict())

    def __getitem__(self, k):
        return self.dict().get(k)

    def __iter__(self):
        return iter(self.dict())


class ExportedModel(Model):
    pass



import shutil
from pathlib import Path
from typing import Generator
from aiofiles import os
from aiofiles.os import wrap
from aiofiles.threadpool import open
from aiofiles.base import AiofilesContextManager

from web_foundation.kernel.messaging.channel import IChannel
from web_foundation.resources.files.interface import FilesResource

rmtree = wrap(shutil.rmtree)


class OsFilesResource(FilesResource):

    def __init__(self, root: Path, *args, **kwargs):
        self.root = root
        super().__init__(*args, **kwargs)

    async def exists(self, path: Path) -> bool: # Todo async
        return self.root.joinpath(path).exists()

    async def get_full_path(self, path: Path) -> Path: # Todo async
        return self.root.joinpath(path)

    async def open(self, path: Path, *args, **kwargs) -> AiofilesContextManager:
        return open(self.root.joinpath(path), *args, **kwargs)

    async def remove(self, path: Path, recursive: bool = False, *args, **kwargs):
        if recursive and path.is_dir():
            await rmtree(path, *args, **kwargs)
        else:
            await os.remove(path, *args, **kwargs)

    async def replace(self, path: Path, new_path: Path, *args, **kwargs):
        await os.replace(path, new_path, *args, **kwargs)

    async def mkdir(self, path: Path, *args, **kwargs):
        await os.mkdir(path, *args, **kwargs)

    async def makedirs(self, path: Path, *args, **kwargs):
        await os.makedirs(path, exist_ok=True)

    async def rename(self, path: Path, new_name: str, *args, **kwargs):
        await os.rename(path, path.parent.joinpath(new_name))

    async def list(self, path: Path, *args, **kwargs) -> Generator[Path, None, None]:
        target = self.root.joinpath(path)
        return target.iterdir()

    async def shutdown(self):
        pass

    async def init(self, channel: IChannel, *args, **kwargs):
        await super(OsFilesResource, self).init(channel, *args, **kwargs)
        if not self.root.exists():
            raise FileNotFoundError(str(self.root))
